#!/usr/bin/env python
# -*- coding:utf-8 -*-

# file:tablesMetadata.py.py
# author:Nathan
# datetime:2021/8/26 14:56
# software: PyCharm

"""
    Get metadata of all tables
"""
from .commons import str_to_all_small, str_to_little_camel_case, str_to_big_camel_case


class TableMetadata(object):
    TYPE_MAPPING = [
        {
            "database": "mysql",
            "data_map": {
                "int": ["int", "bigint", "smallint", "mediumint", "integer", "tinyint"],
                "float": ["float", "double", "real", "numeric", "decimal"]
            }
        },
        {
            "database": "postgresql",
            "data_map": {
                "int": ["int", "bigint", "smallint", "mediumint"],
                "float": ["float", "double", "decimal"],
                "decimal": ["decimal"],
                "str": ["char", "varchar", "tinytext", "text", "mediumtext", "longtext"]
            }
        }
    ]

    @classmethod
    def get_tables_metadata(cls, metadata, reflection_views) -> dict:
        """
            获取数据库数据
            :param metadata: sqlalchemy元数据
            :param reflection_views: 需要反射的视图名称列表
        """

        # Get all tables object
        table_objs = metadata.tables.values()
        table_dict = {}

        # Traverse each table object to get corresponding attributes to form an attribute dictionary
        for table in table_objs:

            table_name = str(table)
            table_dict[table_name] = {}
            table_dict[table_name]['table_name'] = table_name
            table_dict[table_name]['table_name_all_small'] = str_to_all_small(table_name)
            table_dict[table_name]['table_name_little_camel_case'] = str_to_little_camel_case(table_name)
            table_dict[table_name]['table_name_big_camel_case'] = str_to_big_camel_case(table_name)

            # 如果该表是一个视图
            if table_name in reflection_views:
                table_dict[table_name]['is_view'] = True
                table_dict[table_name]['filter_field'] = []

                table_dict[table_name]['columns'] = []
                for column in table.columns.values():
                    temp_column_dict = {
                        "field_name": str(column.name),
                    }

                    for type_ in cls.TYPE_MAPPING:
                        if str(metadata.bind.url).split('+')[0] != type_['database']:
                            continue
                        for python_type, sql_type_list in type_['data_map'].items():
                            if str(column.type).lower() in sql_type_list:
                                temp_column_dict['field_type'] = python_type
                                break

                    temp_column_dict.setdefault('field_type', 'str')
                    table_dict[table_name]['columns'].append(temp_column_dict)

                continue

            table_dict[table_name]['is_view'] = False

            from sqlalchemy.engine import reflection
            insp = reflection.Inspector.from_engine(metadata.bind)
            # 初始化为空列表
            table_dict[table_name]['primary_key_columns'] = insp.get_pk_constraint(table_name)['constrained_columns']
            table_dict[table_name]['columns'] = {}

            # Traverse each columns to get corresponding attributes
            for column in table.columns.values():
                table_dict[table_name]['columns'][str(column.name)] = {}
                table_dict[table_name]['columns'][str(column.name)]['name'] = str(column.name)

                for type_ in cls.TYPE_MAPPING:
                    if str(metadata.bind.url).split('+')[0] != type_['database']:
                        continue
                    for python_type, sql_type_list in type_['data_map'].items():
                        if str(column.type).lower() in sql_type_list:
                            table_dict[table_name]['columns'][str(column.name)]['type'] = python_type
                            break
                table_dict[table_name]['columns'][str(column.name)].setdefault('type', 'str')

                # 是否自动递增
                table_dict[table_name]['columns'][str(column.name)][
                    'is_autoincrement'] = True if column.autoincrement is True else False

                # 是否可以为空
                table_dict[table_name]['columns'][str(column.name)]['nullable'] = column.nullable

                # 是否存在默认值
                table_dict[table_name]['columns'][str(column.name)][
                    'is_exist_default'] = True if column.server_default is not None else False

        return table_dict



"""
Provides missing Postgresql types.

Use GeoAlchemy2 to support Postgis types.
Implement mockups for Point and LTree type, but generated models won't work out
of the box: a better way would be to use a package implementing proper classes
which could be imported in the target project.

"""

from sqlalchemy.dialects.postgresql.base import ischema_names, PGTypeCompiler
from sqlalchemy import types as sqltypes

try:
    import geoalchemy2
except ImportError:
    pass


class LTREE(sqltypes.TypeEngine):
    """Postgresql LTREE type mockup."""

    __visit_name__ = 'LTREE'


class POINT(sqltypes.TypeEngine):
    """Postgresql POINT type mockup."""

    __visit_name__ = 'POINT'


ischema_names['ltree'] = LTREE
ischema_names['point'] = POINT

PGTypeCompiler.visit_LTREE = lambda self, type_: 'LTREE'
PGTypeCompiler.visit_POINT = lambda self, type_: 'POINT'



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# file:checkTable.py
# author:Itsuka
# datetime:2021/8/25 9:53
# software: PyCharm

"""
    检验表是否符合生成规则
"""

import keyword

from .loggings import loggings
from .tablesMetadata import TableMetadata


class CheckTable(object):

    # check the Primary key
    # 检查table主键
    @classmethod
    def check_primary_key(cls, table_dict):
        """
        根据代码生成模式，自动读取所有表或所需表，检验主键后返回合规的表列表
        :return: 符合规范的表名列表，即有且仅有一个主键，没有符合规范的情况下返回None
        """

        available_tables = []
        invalid_tables = []

        for table in table_dict.values():
            if table['is_view']:
                # 是一个视图，不进行检查
                continue
            elif len(table['primary_key_columns']) == 0:
                # 表中没有主键
                invalid_tables.append(table['table_name'])
                loggings.warning(1, 'table {0} do not have a primary key'.format(table['table_name']))
            # elif len(table['primary_key_columns']) > 1:
            #     # 表中有复数个主键
            #     invalid_tables.append(table['table_name'])
            #     loggings.warning(1, 'table {0} has multiple primary keys'.format(table['table_name']))
            else:
                available_tables.append(table['table_name'])

        return available_tables, invalid_tables

    # check keywords of python in tables
    # 检查表名和字段名，是否和Python的关键字冲突
    @classmethod
    def check_keyword_conflict(cls, table_dict):
        """
        check whether the table name or column name is a keyword of python
        :return: True while no table name is a keyword, else return False
        """

        available_table = []
        invalid_table = []

        for table in table_dict.values():

            if table['is_view']:
                # 是一个视图，不进行检查
                continue

            flag = True

            # 检查表名是否为python关键字
            if keyword.iskeyword(table['table_name']):
                loggings.warning(1, 'table "{0}" is a keyword of python'.format(table['table_name']))
                flag = False

            for column in table['columns'].values():
                # 检查表字段是否为python关键字
                if keyword.iskeyword(column['name']):
                    loggings.warning(1, 'column "{0}.{1}" is a keyword of python'.format(table['table_name'],
                                                                                         column['name']))
                    flag = False

            if flag:
                available_table.append(table['table_name'])
            else:
                invalid_table.append(table['table_name'])

        return available_table, invalid_table

    # 入口函数定义
    @classmethod
    def main(cls, metadata, reflection_views):
        """
            建立数据库连接时对表进行检查，筛去没有唯一自增主键、表名/字段名与Python关键字有冲突的表
            :param metadata: 数据库元数据
            :param reflection_views: 需要反射的视图名称列表
        """

        table_dict = TableMetadata.get_tables_metadata(metadata, reflection_views)
        invalid_tables = {}

        # check table primary key
        available_table, invalid_table = cls.check_primary_key(table_dict)
        invalid_tables['primary_key'] = invalid_table
        for invalid in invalid_table:
            table_dict.pop(invalid)

        # check the keyword
        available_table, invalid_table = cls.check_keyword_conflict(table_dict)
        available_tables = available_table
        invalid_tables['keyword'] = invalid_table
        for invalid in invalid_table:
            table_dict.pop(invalid)

        if len(invalid_tables) > 0:
            loggings.warning(
                1,
                "A total of {0} tables check passed."
                "The following {1} tables do not meet the specifications and cannot be generated: {2}."
                    .format(
                        len(available_tables),
                        len(invalid_tables['primary_key'] + invalid_tables['keyword']),
                        ",".join(invalid_tables['primary_key'] + invalid_tables['keyword'])
                    )
            )

            return table_dict, invalid_tables

        loggings.info(1, "All table checks passed, a total of {0} tables.".format(len(available_tables)))

        return table_dict, invalid_tables



""" """
from __future__ import unicode_literals, division, print_function, absolute_import

import argparse
import importlib
import os
import sys

from sqlalchemy.engine import create_engine
from sqlalchemy.schema import MetaData

from . import modelcodegen
from .controllercodegen.codegenerator import CodeGenerator as ControllerCodeGenerator
from .modelcodegen.codegen import CodeGenerator as SQLCodeGenerator
from .utils.tablesMetadata import TableMetadata


def import_dialect_specificities(engine):
    dialect_name = '.' + engine.dialect.name
    try:
        importlib.import_module(dialect_name, 'modelcodegen.dialects')
    except ImportError:
        pass


def main():
    parser = argparse.ArgumentParser(description='Generates SQLAlchemy model code from an existing database.')
    parser.add_argument('url', nargs='?', help='SQLAlchemy url to the database')
    parser.add_argument('--version', action='store_true', help="print the version number and exit")
    parser.add_argument('--schema', help='load tables from an alternate schema')
    parser.add_argument('--tables', help='tables to process (comma-separated, default: all)')
    parser.add_argument('--noviews', action='store_true', help="ignore views")
    parser.add_argument('--noindexes', action='store_true', help='ignore indexes')
    parser.add_argument('--noconstraints', action='store_true', help='ignore constraints')
    parser.add_argument('--nojoined', action='store_true', help="don't autodetect joined table inheritance")
    parser.add_argument('--noinflect', action='store_true', help="don't try to convert tables names to singular form")
    parser.add_argument('--noclasses', action='store_true', help="don't generate classes, only tables")
    parser.add_argument('--notables', action='store_true', help="don't generate tables, only classes")
    parser.add_argument('--outdir', help='file to write output to (default: stdout)')
    parser.add_argument('--models_layer', action='store_true', help='model file to write output to direction models')
    parser.add_argument('--controller_layer', action='store_true',
                        help='controller file to write output to direction controllers')
    parser.add_argument('--nobackrefs', action='store_true', help="don't include backrefs")
    parser.add_argument('--flask', action='store_true', help="use Flask-SQLAlchemy columns")
    parser.add_argument('--ignore-cols',
                        help="Don't check foreign key constraints on specified columns (comma-separated)")
    parser.add_argument('--nocomments', action='store_true', help="don't render column comments")
    args = parser.parse_args()

    if args.version:
        print(modelcodegen.version)
        return
    if not args.url:
        print('You must supply a url\n', file=sys.stderr)
        parser.print_help()
        return
    engine = create_engine(args.url)
    import_dialect_specificities(engine)
    metadata = MetaData(engine)
    tables = args.tables.split(',') if args.tables else None
    ignore_cols = args.ignore_cols.split(',') if args.ignore_cols else None
    metadata.reflect(engine, args.schema, not args.noviews, tables)
    outdir = args.outdir if args.outdir else sys.stdout
    generator = SQLCodeGenerator(metadata, args.noindexes, args.noconstraints,
                                 args.nojoined, args.noinflect, args.nobackrefs,
                                 args.flask, ignore_cols, args.noclasses, args.nocomments, args.notables)

    if args.models_layer:
        os.makedirs(model_dir := os.path.join(outdir, 'models'), exist_ok=True)
        generator.render(model_dir)

    if args.controller_layer:
        os.makedirs(controller_dir := os.path.join(outdir, 'controller'), exist_ok=True)
        reflection_views = [model.table.name for model in generator.models if type(model) == modelcodegen.codegen.ModelTable]
        table_dict = TableMetadata.get_tables_metadata(
            metadata=metadata,
            reflection_views=reflection_views,
        )
        generator = ControllerCodeGenerator(table_dict, args.flask, args.url)
        generator.controller_codegen(controller_dir=controller_dir)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# file:filetemplate.py
# author:Itsuka
# datetime:2021/8/26 10:27
# software: PyCharm

"""
    provide file template here
"""


class FileTemplate(object):

    init_flask_template = """\
#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""

    init_template = """\
#!/usr/bin/env python
# -*- coding:utf-8 -*-

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

SQLALCHEMY_DATABASE_URI = "{databaseUrl}"
engine = create_engine(SQLALCHEMY_DATABASE_URI)
Session = sessionmaker(bind=engine)
session = Session()

"""

    basic_template = """\
#!/usr/bin/env python
# -*- coding:utf-8 -*-

{imports}


class {class_name}({parent_model}):
"""

    flask_add_template = """
    # add
    @classmethod
    def add(cls, **kwargs):
        try:
            model = {parent_model}(
                {column_init}
            )
            db.session.add(model)
            db.session.commit()
            results = {{
                'add_time': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                {add_result_primary_key}
            }}
            return {{'code': "200", 'message': "成功！", 'data': results}}
            
        except Exception as e:
            db.session.rollback()
            return {{'code': "5001", 'message': "数据库错误", 'data': {{'error': str(e)}}}}
        finally:
            db.session.close()
"""

    flask_get_template = """
    # get
    @classmethod
    def get(cls, **kwargs):
        try:
            filter_list = []
            {get_filter_list}
            page = int(kwargs.get('Page', 1))
            size = int(kwargs.get('Size', 10))
            
            {model_lower}_info = db.session.query(cls).filter(*filter_list)
            
            count = {model_lower}_info.count()
            pages = math.ceil(count / size)
            {model_lower}_info = {model_lower}_info.limit(size).offset((page - 1) * size).all()

            results = cls.to_dict({model_lower}_info)
            return {{'code': "200", 'message': "成功！", 'totalCount': count, 'totalPage': pages, 'data': results}}
            
        except Exception as e:
            return {{'code': "5001", 'message': "数据库错误", 'data': {{'error': str(e)}}}}
        finally:
            db.session.close()
"""

    flask_delete_template_physical = """
    # delete
    @classmethod
    def delete(cls, **kwargs):
        try:
            filter_list = []
            {filter_list_init}
            res = db.session.query(cls).filter(*filter_list).with_for_update()
            
            if not res.first():
                return {{'code': "5001", 'message': "数据库错误", 'error': "数据不存在"}}

            results = {{
                'delete_time': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                {results_primary_keys}
            }}
            res.delete()
            db.session.commit()

            return {{'code': "200", 'message': "成功！", 'data': results}}

        except Exception as e:
            db.session.rollback()
            
            return {{'code': "5001", 'message': "数据库错误", 'data': {{'error': str(e)}}}}
        finally:
            db.session.close()
"""

    flask_update_template_physical = """
    # update
    @classmethod
    def update(cls, **kwargs):
        try:
            filter_list = []
            {filter_list_init}
            res = db.session.query(cls).filter(*filter_list).with_for_update()

            if not res.first():
                return {{'code': "5001", 'message': "数据库错误", 'error': "数据不存在"}}
                
            results = {{
                'update_time': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                {results_primary_keys}
            }}
            
            res.update(kwargs)
            db.session.commit()
            
            return {{'code': "200", 'message': "成功！", 'data': results}}

        except Exception as e:
            db.session.rollback()
            
            return {{'code': "5001", 'message': "数据库错误", 'data': {{'error': str(e)}}}}
        finally:
            db.session.close()
"""

    add_template = """
    # add
    @classmethod
    def add(cls, **kwargs):        
        try:
            model = {parent_model}(
                {column_init}
            )
            session.add(model)
            session.commit()
            results = {{
                'add_time': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                {add_result_primary_key}
            }}
            return {{'code': "200", 'message': "成功！", 'data': results}}

        except Exception as e:
            session.rollback()
            
            return {{'code': "5001", 'message': "数据库错误", 'data': {{'error': str(e)}}}}
        finally:
            session.close()
"""

    get_template = """
    # get
    @classmethod
    def get(cls, **kwargs):        
        try:
            filter_list = []
            {get_filter_list}
            page = int(kwargs.get('Page', 1))
            size = int(kwargs.get('Size', 10))

            {model_lower}_info = session.query(cls).filter(*filter_list)

            count = {model_lower}_info.count()
            pages = math.ceil(count / size)
            {model_lower}_info = {model_lower}_info.limit(size).offset((page - 1) * size).all()

            results = cls.to_dict({model_lower}_info)
            return {{'code': "200", 'message': "成功！", 'totalCount': count, 'totalPage': pages, 'data': results}}

        except Exception as e:
            return {{'code': "5001", 'message': "数据库错误", 'data': {{'error': str(e)}}}}
        finally:
            session.close()
"""

    delete_template_physical = """
    # delete
    @classmethod
    def delete(cls, **kwargs):        
        try:
            filter_list = []
            {filter_list_init}
            res = session.query(cls).filter(*filter_list).with_for_update()

            if not res.first():
                return {{'code': "5001", 'message': "数据库错误", 'error': "数据不存在"}}
                
            results = {{
                'delete_time': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                {results_primary_keys}
            }}
            res.delete()
            session.commit()

            return {{'code': "200", 'message': "成功！", 'data': results}}

        except Exception as e:
            session.rollback()
            
            return {{'code': "5001", 'message': "数据库错误", 'data': {{'error': str(e)}}}}
        finally:
            session.close()
"""

    update_template_physical = """
    # update
    @classmethod
    def update(cls, **kwargs):        
        try:
            filter_list = []
            {filter_list_init}
            res = session.query(cls).filter(*filter_list).with_for_update()

            if not res.first():
                return {{'code': "5001", 'message': "数据库错误", 'error': "数据不存在"}}
                
            results = {{
                'update_time': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                {results_primary_keys}
            }}

            res.update(kwargs)
            session.commit()

            return {{'code': "200", 'message': "成功！", 'data': results}}

        except Exception as e:
            session.rollback()
            return {{'code': "5001", 'message': "数据库错误", 'data': {{'error': str(e)}}}}
        finally:
            session.close()
"""


"""Contains the code generation logic and helper functions."""

from __future__ import unicode_literals, division, print_function, absolute_import

import inspect
import os.path
import re
import sys
from collections import defaultdict
from inspect import ArgSpec
from keyword import iskeyword

import sqlalchemy
from sqlalchemy import (Enum, ForeignKeyConstraint, PrimaryKeyConstraint, CheckConstraint, UniqueConstraint, Table,
                        Column)
from sqlalchemy.schema import ForeignKey
from sqlalchemy.types import Boolean, String
from sqlalchemy.util import OrderedDict

try:
    from sqlalchemy.sql.expression import text, TextClause
except ImportError:
    # SQLAlchemy < 0.8
    from sqlalchemy.sql.expression import text, _TextClause as TextClause

_re_boolean_check_constraint = re.compile(r"(?:(?:.*?)\.)?(.*?) IN \(0, 1\)")
_re_column_name = re.compile(r'(?:(["`]?)(?:.*)\1\.)?(["`]?)(.*)\2')
_re_enum_check_constraint = re.compile(r"(?:(?:.*?)\.)?(.*?) IN \((.+)\)")
_re_enum_item = re.compile(r"'(.*?)(?<!\\)'")
_re_invalid_identifier = re.compile(r'[^a-zA-Z0-9_]' if sys.version_info[0] < 3 else r'(?u)\W')

_re_first_cap = re.compile('(.)([A-Z][a-z]+)')
_re_all_cap = re.compile('([a-z0-9])([A-Z])')

_flask_prepend = 'db.'


class _DummyInflectEngine(object):
    def singular_noun(self, noun):
        return noun
    def plural_noun(self, noun):  # needed for backrefs
        import inflect
        inflect_engine = inflect.engine()
        return inflect_engine.plural_noun(noun)


# In SQLAlchemy 0.x, constraint.columns is sometimes a list, on 1.x onwards, always a ColumnCollection
def _get_column_names(constraint):
    if isinstance(constraint.columns, list):
        return constraint.columns
    return list(constraint.columns.keys())


def _convert_to_valid_identifier(name):
    assert name, 'Identifier cannot be empty'
    if name[0].isdigit() or iskeyword(name):
        name = '_' + name
    return _re_invalid_identifier.sub('_', name)


def _get_compiled_expression(statement):
    """Returns the statement in a form where any placeholders have been filled in."""
    if isinstance(statement, TextClause):
        return statement.text

    dialect = statement._from_objects[0].bind.dialect
    compiler = statement._compiler(dialect)

    # Adapted from http://stackoverflow.com/a/5698357/242021
    class LiteralCompiler(compiler.__class__):
        def visit_bindparam(self, bindparam, within_columns_clause=False, literal_binds=False, **kwargs):
            return super(LiteralCompiler, self).render_literal_bindparam(
                bindparam, within_columns_clause=within_columns_clause,
                literal_binds=literal_binds, **kwargs
            )

    compiler = LiteralCompiler(dialect, statement)
    return compiler.process(statement)


def _get_constraint_sort_key(constraint):
    if isinstance(constraint, CheckConstraint):
        return 'C{0}'.format(constraint.sqltext)
    return constraint.__class__.__name__[0] + repr(_get_column_names(constraint))


def _get_common_fk_constraints(table1, table2):
    """Returns a set of foreign key constraints the two tables have against each other."""
    c1 = set(c for c in table1.constraints if isinstance(c, ForeignKeyConstraint) and
             c.elements[0].column.table == table2)
    c2 = set(c for c in table2.constraints if isinstance(c, ForeignKeyConstraint) and
             c.elements[0].column.table == table1)
    return c1.union(c2)


def _getargspec_init(method):
    try:
        return inspect.getargspec(method)
    except TypeError:
        if method is object.__init__:
            return ArgSpec(['self'], None, None, None)
        else:
            return ArgSpec(['self'], 'args', 'kwargs', None)


def _render_column_type(coltype):
    args = []
    if isinstance(coltype, Enum):
        args.extend(repr(arg) for arg in coltype.enums)
        if coltype.name is not None:
            args.append('name={0!r}'.format(coltype.name))
    else:
        # All other types
        argspec = _getargspec_init(coltype.__class__.__init__)
        defaults = dict(zip(argspec.args[-len(argspec.defaults or ()):], argspec.defaults or ()))
        missing = object()
        use_kwargs = False
        for attr in argspec.args[1:]:
            # Remove annoyances like _warn_on_bytestring
            if attr.startswith('_'):
                continue

            value = getattr(coltype, attr, missing)
            default = defaults.get(attr, missing)
            if value is missing or value == default:
                use_kwargs = True
            elif use_kwargs:
                args.append('{0}={1}'.format(attr, repr(value)))
            else:
                args.append(repr(value))

    text = _flask_prepend + coltype.__class__.__name__

    # In case of ENUM from sqlalchemy.dialects, the flask used db.Enum
    if text == "db.ENUM":
        text = "db.Enum"

    if args:
        text += '({0})'.format(', '.join(args))

    return text


def _render_column(column, show_name):
    kwarg = []
    is_sole_pk = column.primary_key and len(column.table.primary_key) == 1
    dedicated_fks = [c for c in column.foreign_keys if len(c.constraint.columns) == 1]
    is_unique = any(isinstance(c, UniqueConstraint) and set(c.columns) == set([column])
                    for c in column.table.constraints)
    is_unique = is_unique or any(i.unique and set(i.columns) == set([column]) for i in column.table.indexes)
    has_index = any(set(i.columns) == set([column]) for i in column.table.indexes)

    # Render the column type if there are no foreign keys on it or any of them points back to itself
    render_coltype = not dedicated_fks or any(fk.column is column for fk in dedicated_fks)

    if column.key != column.name:
        kwarg.append('key')
    if column.primary_key:
        kwarg.append('primary_key')
    if not column.nullable and not is_sole_pk:
        kwarg.append('nullable')
    if is_unique:
        column.unique = True
        kwarg.append('unique')
    elif has_index:
        column.index = True
        kwarg.append('index')
    if column.server_default:
        server_default = 'server_default=' + _flask_prepend + 'FetchedValue()'

    comment = getattr(column, 'comment', None)
    return _flask_prepend + 'Column({0})'.format(', '.join(
        ([repr(column.name)] if show_name else []) +
        ([_render_column_type(column.type)] if render_coltype else []) +
        [_render_constraint(x) for x in dedicated_fks] +
        [repr(x) for x in column.constraints] +
        ['{0}={1}'.format(k, repr(getattr(column, k))) for k in kwarg] +
        ([server_default] if column.server_default else []) +
        (['info={!r}'.format(comment)] if comment else [])
    ))


def _render_constraint(constraint):
    def render_fk_options(*opts):
        opts = [repr(opt) for opt in opts]
        for attr in 'ondelete', 'onupdate', 'deferrable', 'initially', 'match':
            value = getattr(constraint, attr, None)
            if value:
                opts.append('{0}={1!r}'.format(attr, value))

        return ', '.join(opts)

    if isinstance(constraint, ForeignKey):
        remote_column = '{0}.{1}'.format(constraint.column.table.fullname, constraint.column.name)
        return _flask_prepend + 'ForeignKey({0})'.format(render_fk_options(remote_column))
    elif isinstance(constraint, ForeignKeyConstraint):
        local_columns = _get_column_names(constraint)
        remote_columns = ['{0}.{1}'.format(fk.column.table.fullname, fk.column.name)
                          for fk in constraint.elements]
        return _flask_prepend + 'ForeignKeyConstraint({0})'.format(render_fk_options(local_columns, remote_columns))
    elif isinstance(constraint, CheckConstraint):
        return _flask_prepend + 'CheckConstraint({0!r})'.format(_get_compiled_expression(constraint.sqltext))
    elif isinstance(constraint, UniqueConstraint):
        columns = [repr(col.name) for col in constraint.columns]
        return _flask_prepend + 'UniqueConstraint({0})'.format(', '.join(columns))


def _underscore(name):
    """Converts CamelCase to camel_case. See http://stackoverflow.com/questions/1175208"""
    s1 = _re_first_cap.sub(r'\1_\2', name)
    return _re_all_cap.sub(r'\1_\2', s1).lower()


def _is_model_descendant(model_a, model_b):
    """Check to see if model class A inherits from another model class B"""
    if model_a.name == model_b.name:
        return True
    if not model_b.children:
        return False
    return any(_is_model_descendant(model_a, b) for b in model_b.children)


def _render_index(index):
    columns = [repr(col.name) for col in index.columns]
    return _flask_prepend + 'Index({0!r}, {1})'.format(index.name, ', '.join(columns))


class ImportCollector(OrderedDict):
    def add_import(self, obj):
        type_ = type(obj) if not isinstance(obj, type) else obj
        pkgname = 'sqlalchemy' if type_.__name__ in sqlalchemy.__all__ else type_.__module__  # @UndefinedVariable
        self.add_literal_import(pkgname, type_.__name__)

    def add_literal_import(self, pkgname, name):
        names = self.setdefault(pkgname, set())
        names.add(name)

    def render(self):
        return '\n'.join('from {0} import {1}'.format(package, ', '.join(sorted(names)))
                         for package, names in self.items())

    def render_foreign_object(self, object_list: list):
        text = ""
        for object_ in object_list:
            text += "\n"+"from .{0} import {1}".format(object_[0].lower()+object_[1:], object_)
        return text

class Model(object):
    def __init__(self, table):
        super(Model, self).__init__()
        self.table = table
        self.schema = table.schema

        # Adapt column types to the most reasonable generic types (ie. VARCHAR -> String)
        for column in table.columns:
            cls = column.type.__class__
            for supercls in cls.__mro__:
                if hasattr(supercls, '__visit_name__'):
                    cls = supercls
                if supercls.__name__ != supercls.__name__.upper() and not supercls.__name__.startswith('_'):
                    break

            column.type = column.type.adapt(cls)

    def add_imports(self, collector):
        if self.table.columns:
            collector.add_import(Column)

        for column in self.table.columns:
            collector.add_import(column.type)
            if column.server_default:
                collector.add_literal_import('sqlalchemy.schema', 'FetchedValue')

        for constraint in sorted(self.table.constraints, key=_get_constraint_sort_key):
            if isinstance(constraint, ForeignKeyConstraint):
                if len(constraint.columns) > 1:
                    collector.add_literal_import('sqlalchemy', 'ForeignKeyConstraint')
                else:
                    collector.add_literal_import('sqlalchemy', 'ForeignKey')
            elif isinstance(constraint, UniqueConstraint):
                if len(constraint.columns) > 1:
                    collector.add_literal_import('sqlalchemy', 'UniqueConstraint')
            elif not isinstance(constraint, PrimaryKeyConstraint):
                collector.add_import(constraint)

        for index in self.table.indexes:
            if len(index.columns) > 1:
                collector.add_import(index)


class ModelTable(Model):
    def add_imports(self, collector):
        super(ModelTable, self).add_imports(collector)
        collector.add_import(Table)

    def render(self):
        met = ' metadata,' if _flask_prepend == '' else ''
        text = 't_{0} = {1}Table(\n    {0!r},{2}\n'.format(self.table.name, _flask_prepend, met)

        for column in self.table.columns:
            text += '    {0},\n'.format(_render_column(column, True))

        for constraint in sorted(self.table.constraints, key=_get_constraint_sort_key):
            if isinstance(constraint, PrimaryKeyConstraint):
                continue
            if isinstance(constraint, (ForeignKeyConstraint, UniqueConstraint)) and len(constraint.columns) == 1:
                continue
            text += '    {0},\n'.format(_render_constraint(constraint))

        for index in self.table.indexes:
            if len(index.columns) > 1:
                text += '    {0},\n'.format(_render_index(index))

        if self.schema:
            text += "    schema='{0}',\n".format(self.schema)

        return text.rstrip('\n,') + '\n)'


class ModelClass(Model):
    parent_name = 'Base'

    def __init__(self, table, association_tables, inflect_engine, detect_joined):
        super(ModelClass, self).__init__(table)
        self.name = self._tablename_to_classname(table.name)
        self.children = []
        self.attributes = OrderedDict()

        # Assign attribute names for columns
        for column in table.columns:
            self._add_attribute(column.name, column)

        # Add many-to-one relationships
        pk_column_names = set(col.name for col in table.primary_key.columns)
        for constraint in sorted(table.constraints, key=_get_constraint_sort_key):
            if isinstance(constraint, ForeignKeyConstraint):
                target_cls = self._tablename_to_classname(constraint.elements[0].column.table.name)
                if (detect_joined and self.parent_name == 'Base' and
                        set(_get_column_names(constraint)) == pk_column_names):
                    self.parent_name = target_cls
                else:
                    relationship_ = ManyToOneRelationship(self.name, target_cls, constraint, inflect_engine)
                    self._add_attribute(relationship_.preferred_name, relationship_)

        # Add many-to-many relationships
        for association_table in association_tables:
            fk_constraints = [c for c in association_table.constraints if isinstance(c, ForeignKeyConstraint)]
            fk_constraints.sort(key=_get_constraint_sort_key)
            target_cls = self._tablename_to_classname(fk_constraints[1].elements[0].column.table.name)
            relationship_ = ManyToManyRelationship(self.name, target_cls, association_table, inflect_engine)
            self._add_attribute(relationship_.preferred_name, relationship_)

    @staticmethod
    def _tablename_to_classname(tablename):
        camel_case_name = ''.join(part[:1].upper() + part[1:] for part in re.split(r'_|-', tablename))
        # return inflect_engine.singular_noun(camel_case_name) or camel_case_name
        return camel_case_name

    def _add_attribute(self, attrname, value):
        attrname = tempname = _convert_to_valid_identifier(attrname)
        counter = 1
        while tempname in self.attributes:
            tempname = attrname + str(counter)
            counter += 1

        self.attributes[tempname] = value
        return tempname

    def add_imports(self, collector):
        super(ModelClass, self).add_imports(collector)

        if any(isinstance(value, Relationship) for value in self.attributes.values()):
            collector.add_literal_import('sqlalchemy.orm', 'relationship')

        for child in self.children:
            child.add_imports(collector)

    def render(self):
        text = 'class {0}({1}):\n'.format(self.name, self.parent_name)
        text += '    __tablename__ = {0!r}\n'.format(self.table.name)

        # Render constraints and indexes as __table_args__
        table_args = []
        for constraint in sorted(self.table.constraints, key=_get_constraint_sort_key):
            if isinstance(constraint, PrimaryKeyConstraint):
                continue
            if isinstance(constraint, (ForeignKeyConstraint, UniqueConstraint)) and len(constraint.columns) == 1:
                continue
            table_args.append(_render_constraint(constraint))
        for index in self.table.indexes:
            if len(index.columns) > 1:
                table_args.append(_render_index(index))

        table_kwargs = {}
        if self.schema:
            table_kwargs['schema'] = self.schema

        kwargs_items = ', '.join('{0!r}: {1!r}'.format(key, table_kwargs[key]) for key in table_kwargs)
        kwargs_items = '{{{0}}}'.format(kwargs_items) if kwargs_items else None
        if table_kwargs and not table_args:
            text += '    __table_args__ = {0}\n'.format(kwargs_items)
        elif table_args:
            if kwargs_items:
                table_args.append(kwargs_items)
            if len(table_args) == 1:
                table_args[0] += ','
            text += '    __table_args__ = (\n        {0}\n    )\n'.format(',\n        '.join(table_args))

        # Render columns
        text += '\n'
        for attr, column in self.attributes.items():
            if isinstance(column, Column):
                show_name = attr != column.name
                text += '    {0} = {1}\n'.format(attr, _render_column(column, show_name))

        # Render relationships
        if any(isinstance(value, Relationship) for value in self.attributes.values()):
            text += '\n'
        for attr, relationship in self.attributes.items():
            if isinstance(relationship, Relationship):
                text += '    {0} = {1}\n'.format(attr, relationship.render())

        # Render subclasses
        for child_class in self.children:
            text += '\n\n' + child_class.render()

        return text


class Relationship(object):
    def __init__(self, source_cls, target_cls):
        super(Relationship, self).__init__()
        self.source_cls = source_cls
        self.target_cls = target_cls
        self.kwargs = OrderedDict()
        self.backref_name = _underscore(self.source_cls)

    def render(self):
        text = _flask_prepend + 'relationship('
        args = [repr(self.target_cls)]

        if 'secondaryjoin' in self.kwargs:
            text += '\n        '
            delimiter, end = ',\n        ', '\n    )'
        else:
            delimiter, end = ', ', ')'

        args.extend([key + '=' + value for key, value in self.kwargs.items()])
        return text + delimiter.join(args) + end

    def make_backref(self, relationships, classes):
        backref = self.backref_name
        original_backref = backref
        # Check if backref already exists for relationship source_cls to target_cls and add suffix
        suffix = 0
        while (self.target_cls, backref) in [(x.target_cls, x.backref_name) for x in relationships]:
            backref = original_backref + str('_{0}'.format(suffix))
            suffix += 1

        self.kwargs['backref'] = repr(backref)
        # Check if any of the target_cls inherit from other target_cls
        # If so, modify backref name of descendant
        # "backref({0}, lazy='dynamic')".format(repr(backref))
        for rel in [x for x in relationships if 'backref' in x.kwargs]:
            if self.target_cls in classes and rel.target_cls in classes:
                if _is_model_descendant(classes[self.target_cls], classes[rel.target_cls]):
                    self.backref_name = self.target_cls.lower() + '_' + backref
                    self.kwargs['backref'] = repr(self.backref_name)
                if _is_model_descendant(classes[rel.target_cls], classes[self.target_cls]):
                    backref = rel.backref_name
                    rel.backref_name = rel.target_cls.lower() + '_' + backref
                    rel.kwargs['backref'] = repr(rel.backref_name)


class ManyToOneRelationship(Relationship):
    def __init__(self, source_cls, target_cls, constraint, inflect_engine):
        super(ManyToOneRelationship, self).__init__(source_cls, target_cls)

        column_names = _get_column_names(constraint)
        colname = column_names[0]
        tablename = constraint.elements[0].column.table.name
        if not colname.endswith('_id'):
            self.preferred_name = inflect_engine.singular_noun(tablename) or tablename
        else:
            self.preferred_name = colname[:-3]
        self.backref_name = inflect_engine.plural_noun(self.backref_name)

        # Add uselist=False to One-to-One relationships
        if any(isinstance(c, (PrimaryKeyConstraint, UniqueConstraint)) and
               set(col.name for col in c.columns) == set(column_names)
               for c in constraint.table.constraints):
            self.kwargs['uselist'] = 'False'

        # Handle self referential relationships
        if source_cls == target_cls:
            self.preferred_name = 'parent' if not colname.endswith('_id') else colname[:-3]
            pk_col_names = [col.name for col in constraint.table.primary_key]
            self.kwargs['remote_side'] = '[{0}]'.format(', '.join(pk_col_names))

        # If the two tables share more than one foreign key constraint,
        # SQLAlchemy needs an explicit primaryjoin to figure out which column(s) to join with
        # common_fk_constraints = _get_common_fk_constraints(constraint.table, constraint.elements[0].column.table)
        # if len(common_fk_constraints) > 1:
        # self.kwargs['primaryjoin'] = "'{0}.{1} == {2}.{3}'".format(source_cls, constraint.columns[0], target_cls, constraint.elements[0].column.name)
        if len(constraint.elements) > 1:  #  or 
            self.kwargs['primaryjoin'] = "'and_({0})'".format(', '.join(['{0}.{1} == {2}.{3}'.format(source_cls, k.parent.name, target_cls, k.column.name)
                        for k in constraint.elements]))
        else:
            self.kwargs['primaryjoin'] = "'{0}.{1} == {2}.{3}'".format(source_cls, column_names[0], target_cls,
                                                                       constraint.elements[0].column.name)


class ManyToManyRelationship(Relationship):
    def __init__(self, source_cls, target_cls, assocation_table, inflect_engine):
        super(ManyToManyRelationship, self).__init__(source_cls, target_cls)
        prefix = assocation_table.schema + '.' if assocation_table.schema is not None else ''
        self.kwargs['secondary'] = repr(prefix + assocation_table.name)
        constraints = [c for c in assocation_table.constraints if isinstance(c, ForeignKeyConstraint)]
        constraints.sort(key=_get_constraint_sort_key)
        colname = _get_column_names(constraints[1])[0]
        tablename = constraints[1].elements[0].column.table.name
        self.preferred_name = tablename if not colname.endswith('_id') else colname[:-3] + 's'
        self.backref_name = inflect_engine.plural_noun(self.backref_name)

        # Handle self referential relationships
        if source_cls == target_cls:
            self.preferred_name = 'parents' if not colname.endswith('_id') else colname[:-3] + 's'
            pri_pairs = zip(_get_column_names(constraints[0]), constraints[0].elements)
            sec_pairs = zip(_get_column_names(constraints[1]), constraints[1].elements)
            pri_joins = ['{0}.{1} == {2}.c.{3}'.format(source_cls, elem.column.name, assocation_table.name, col)
                         for col, elem in pri_pairs]
            sec_joins = ['{0}.{1} == {2}.c.{3}'.format(target_cls, elem.column.name, assocation_table.name, col)
                         for col, elem in sec_pairs]
            self.kwargs['primaryjoin'] = (
                repr('and_({0})'.format(', '.join(pri_joins))) if len(pri_joins) > 1 else repr(pri_joins[0]))
            self.kwargs['secondaryjoin'] = (
                repr('and_({0})'.format(', '.join(sec_joins))) if len(sec_joins) > 1 else repr(sec_joins[0]))


class CodeGenerator(object):
    header = '# coding: utf-8'
    footer = ''

    def __init__(self, metadata, noindexes=False, noconstraints=False,
                 nojoined=False, noinflect=False, nobackrefs=False,
                 flask=False, ignore_cols=None, noclasses=False, nocomments=False, notables=False):
        super(CodeGenerator, self).__init__()

        if noinflect:
            inflect_engine = _DummyInflectEngine()
        else:
            import inflect
            inflect_engine = inflect.engine()

        # exclude these column names from consideration when generating association tables
        _ignore_columns = ignore_cols or []
        
        self.flask = flask
        if not self.flask:
            global _flask_prepend
            _flask_prepend = ''

        self.nocomments = nocomments

        # Pick association tables from the metadata into their own set, don't process them normally
        links = defaultdict(lambda: [])
        association_tables = set()
        for table in metadata.tables.values():
            # Link tables have exactly two foreign key constraints and all columns are involved in them
            # except for special columns like id, inserted, and updated
            fk_constraints = [constr for constr in table.constraints if isinstance(constr, ForeignKeyConstraint)]
            if len(fk_constraints) == 2 and all(col.foreign_keys for col in table.columns if col.name not in _ignore_columns):
                association_tables.add(table.name)
                tablename = sorted(fk_constraints, key=_get_constraint_sort_key)[0].elements[0].column.table.name
                links[tablename].append(table)

        # Iterate through the tables and create model classes when possible
        self.models = []
        self.collector = ImportCollector()
        classes = {}
        for table in sorted(metadata.tables.values(), key=lambda t: (t.schema or '', t.name)):
            # Support for Alembic and sqlalchemy-migrate -- never expose the schema version tables
            if table.name in ('alembic_version', 'migrate_version'):
                continue

            if noindexes:
                table.indexes.clear()

            if noconstraints:
                table.constraints = set([table.primary_key])
                table.foreign_keys.clear()
                for col in table.columns:
                    col.foreign_keys.clear()
            else:
                # Detect check constraints for boolean and enum columns
                for constraint in table.constraints.copy():
                    if isinstance(constraint, CheckConstraint):
                        sqltext = _get_compiled_expression(constraint.sqltext)

                        # Turn any integer-like column with a CheckConstraint like "column IN (0, 1)" into a Boolean
                        match = _re_boolean_check_constraint.match(sqltext)
                        if match:
                            colname = _re_column_name.match(match.group(1)).group(3)
                            table.constraints.remove(constraint)
                            table.c[colname].type = Boolean()
                            continue

                        # Turn any string-type column with a CheckConstraint like "column IN (...)" into an Enum
                        match = _re_enum_check_constraint.match(sqltext)
                        if match:
                            colname = _re_column_name.match(match.group(1)).group(3)
                            items = match.group(2)
                            if isinstance(table.c[colname].type, String):
                                table.constraints.remove(constraint)
                                if not isinstance(table.c[colname].type, Enum):
                                    options = _re_enum_item.findall(items)
                                    table.c[colname].type = Enum(*options, native_enum=False)
                                continue

            # Only generate classes when notables is set to True
            if notables:
                model = ModelClass(table, links[table.name], inflect_engine, not nojoined)
                classes[model.name] = model
            elif not table.primary_key or table.name in association_tables or noclasses:
                # Only form model classes for tables that have a primary key and are not association tables
                model = ModelTable(table)
            elif not noclasses:
                model = ModelClass(table, links[table.name], inflect_engine, not nojoined)
                classes[model.name] = model

            self.models.append(model)

            # collect imports for models only if flask is not specified
            if not self.flask:
                model.add_imports(self.collector)

        # Nest inherited classes in their superclasses to ensure proper ordering
        for model in classes.values():
            if model.parent_name != 'Base':
                classes[model.parent_name].children.append(model)
                self.models.remove(model)

        # If backrefs are allowed. Resolve any relationships conflicts where one
        # target class might inherit from another
        if not nobackrefs:
            for model in classes.values():
                visited = []
                for relationship in model.attributes.values():
                    if isinstance(relationship, Relationship):
                        relationship.make_backref(visited, classes)
                        visited.append(relationship)

        if self.flask:
            # Add Flask-SQLAlchemy support
            self.collector.add_literal_import('flask_sqlalchemy', 'SQLAlchemy')
            parent_name = 'BaseModel'

            for model in classes.values():
                if model.parent_name == 'Base':
                    model.parent_name = parent_name
        # else:
        #     self.collector.add_literal_import('sqlalchemy.ext.declarative', 'declarative_base')
        #     self.collector.add_literal_import('sqlalchemy', 'MetaData')

    def render(self, outdir):
        if not os.path.exists(outdir):
            os.mkdir(outdir)

        # Render the model tables and classes
        with open(os.path.join(outdir, f"__init__.py"), "w", encoding="utf-8") as outfile:
            if self.flask:
                s = """\
#!/usr/bin/env python
# -*- coding:utf-8 -*-

from sqlalchemy import inspect
from sqlalchemy.engine import Row
from flask_sqlalchemy import SQLAlchemy

db = SQLAlchemy()


class BaseModel(db.Model):
    __abstract__ = True

    @classmethod
    def to_dict(cls, models):
        # 使用.all()查询的情况
        if isinstance(models, list):
            # 查询结果为空
            if len(models) == 0:
                return []
            # 查询结果不包含所有字段的情况
            if isinstance(models[0], Row):
                return [row._asdict() for row in models]
            # 查询结果包含所有字段的情况
            else:
                return [cls._asdict(model) for model in models]
        # 使用.first()查询的情况
        else:
            if not models:
                return {}
            if isinstance(models, Row):
                return models._asdict()
            return cls._asdict(models)

    def _asdict(self):
        return {c.key: getattr(self, c.key) for c in inspect(self).mapper.column_attrs}
"""
            else:
                s = """\
#!/usr/bin/env python
# -*- coding:utf-8 -*-

from sqlalchemy import inspect
from sqlalchemy.engine import Row
from sqlalchemy.ext.declarative import declarative_base

DeclarativeBase = declarative_base()


class Base(DeclarativeBase):
    __abstract__ = True

    @classmethod
    def to_dict(cls, models):
        # 使用.all()查询的情况
        if isinstance(models, list):
            # 查询结果为空
            if len(models) == 0:
                return []
            # 查询结果不包含所有字段的情况
            if isinstance(models[0], Row):
                return [row._asdict() for row in models]
            # 查询结果包含所有字段的情况
            else:
                return [cls._asdict(model) for model in models]
        # 使用.first()查询的情况
        else:
            if not models:
                return {}
            if isinstance(models, Row):
                return models._asdict()
            return cls._asdict(models)

    def _asdict(self):
        return {c.key: getattr(self, c.key) for c in inspect(self).mapper.column_attrs}
"""
            outfile.write(s)

        for model in self.models:
            outfilename = os.path.join(outdir, f"{model.table.name}.py")
            with open(outfilename, "w", encoding="utf-8") as outfile:
                print(self.header, file=outfile)
                if self.flask:
                    print('from . import db, BaseModel', file=outfile)
                else:
                    # Render the collected imports
                    print(self.collector.render() + '\n\n', file=outfile)

                    if any(isinstance(model, ModelClass) for model in self.models):
                        print('from . import Base', file=outfile)
                    else:
                        print('metadata = MetaData()', file=outfile)

                relationship_object = []
                for attr, relationship in model.attributes.items():
                    if isinstance(relationship, Relationship):
                        relationship_object.append(relationship.target_cls)
                if len(relationship_object) != 0:
                    print(self.collector.render_foreign_object(relationship_object) + '\n\n', file=outfile)

                print('\n', file=outfile)
                print(model.render().rstrip('\n'), file=outfile)

                if self.footer:
                    print(self.footer, file=outfile)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# file:python_sqla_codegen
# author:PigKnight
# datetime:2021/8/21 11:28
# software: PyCharm

"""
    General method
"""

# 连字符转驼峰
import os
import sys


def str_format_convert(string):
    new_string = ''
    for word in string.split('_'):
        if new_string:
            new_string += word.lower().capitalize()
        else:
            new_string = word
    return new_string


# 字符串转全小写
def str_to_all_small(string):
    new_string = string.replace('_', '').lower()

    return new_string


# 字符串转小驼峰
def str_to_little_camel_case(string):
    new_string = ''
    for word in string.split('_'):
        if new_string:
            new_string += word[0].upper() + word[1:]
        else:
            new_string = word

    return new_string


# 字符串转大驼峰
def str_to_big_camel_case(string):
    new_string = ''
    for word in string.split('_'):
        new_string += word[0].upper() + word[1:]

    return new_string


def cur_file_dir():
    '''
    用于找到当前文件的目录

    :return:返回一个绝对路径
    '''
    # 获取脚本路径
    path = sys.path[0]
    # 判断为脚本文件还是py2exe编译后的文件，如果是脚本文件，则返回的是脚本的目录，如果是py2exe编译后的文件，则返回的是编译后的文件路径
    if os.path.isdir(path):
        return path
    elif os.path.isfile(path):
        return os.path.dirname(path)


if __name__ == '__main__':
    ll = ['user_info', 'userInfo', 'userinfo']
    for l in ll:
        print(l, '->', str_to_little_camel_case(l))



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# file:codegenerator.py
# author:Itsuka
# datetime:2021/8/24 10:04
# software: PyCharm

"""
    generate controller layer code
    This generator is a very simple boilerplate for generate controller code with Flask, flask-restful,
    marshmallow, SQLAlchemy and jwt.
    It comes with basic project structure and configuration, including blueprints, application factory
    and basics unit tests.
"""

import os.path
import shutil

from .template.codeblocktemplate import CodeBlockTemplate
from .template.filetemplate import FileTemplate


class CodeGenerator(object):

    def __init__(self, table_dict, flask, databaseUrl):
        super().__init__()
        self.table_dict = table_dict
        self.flask = flask
        self.databaseUrl = databaseUrl

    def controller_codegen(self, controller_dir):

        try:
            codes = {}
            # get table dict
            table_dict = self.table_dict

            # generate code and save in 'codes'
            for table in [table_view for table_view in table_dict.values() if not table_view.get('is_view')]:
                little_camel_case_str = table['table_name_little_camel_case']
                big_camel_case_str = table['table_name_big_camel_case']
                model_name = little_camel_case_str + 'Model'
                class_name = big_camel_case_str + 'Controller'
                parent_model = big_camel_case_str
                primary_key = table['primary_key_columns']

                # combine imports
                if self.flask:
                    imports = CodeBlockTemplate.flask_imports.format(
                        model_name=table['table_name'],
                        parent_model=parent_model
                    )
                    basic = FileTemplate.basic_template.format(
                        imports=imports,
                        class_name=class_name,
                        parent_model=parent_model
                    )
                else:
                    imports = CodeBlockTemplate.imports.format(
                        model_name=table['table_name'],
                        parent_model=parent_model
                    )
                    basic = FileTemplate.basic_template.format(
                        imports=imports,
                        class_name=class_name,
                        parent_model=parent_model
                    )

                # 添加模块
                column_init = ''
                add_result_primary_key = ''
                for column in table['columns'].values():
                    text = CodeBlockTemplate.add_column_init.format(column=column['name'])

                    column_init += text

                # 拼接返回字段
                for each_primary_key in primary_key:
                    add_result_primary_key += CodeBlockTemplate.add_result_primary_key.format(
                        primary_key=each_primary_key
                    )
                if self.flask:
                    add = FileTemplate.flask_add_template.format(
                        parent_model=parent_model,
                        column_init=column_init,
                        add_result_primary_key=add_result_primary_key
                    )
                else:
                    add = FileTemplate.add_template.format(
                        parent_model=parent_model,
                        column_init=column_init,
                        add_result_primary_key=add_result_primary_key
                    )

                # 查询模块
                get_filter_list = ''
                for column in table['columns'].values():
                    # 属于复合主键
                    if column['type'] in ['int', 'float']:
                        # column type is a number
                        text = CodeBlockTemplate.multi_get_filter_num.format(column=column['name'])
                    else:
                        # column type is a string
                        text = CodeBlockTemplate.multi_get_filter_str.format(column=column['name'])

                    get_filter_list += text

                # 属于复合主键
                if self.flask:
                    get = FileTemplate.flask_get_template.format(
                        get_filter_list=get_filter_list,
                        model_lower=table['table_name']
                    )
                else:
                    get = FileTemplate.get_template.format(
                        get_filter_list=get_filter_list,
                        model_lower=table['table_name']
                    )

                # 删除模块
                # 拼接删除方法中的filter和results
                filter_list_init = ''
                results_primary_keys = ''
                for each_primary_key in primary_key:
                    filter_list_init += CodeBlockTemplate.multi_primary_key_filter.format(
                        primary_key=each_primary_key
                    )
                    results_primary_keys += CodeBlockTemplate.multi_primary_key_result.format(
                        primary_key=each_primary_key
                    )
                if self.flask:
                    delete = FileTemplate.flask_delete_template_physical.format(
                        filter_list_init=filter_list_init,
                        results_primary_keys=results_primary_keys,
                    )
                else:
                    delete = FileTemplate.delete_template_physical.format(
                        filter_list_init=filter_list_init,
                        results_primary_keys=results_primary_keys,
                    )

                # 更新模块

                # 拼接更新方法中的filter_list_init
                filter_list_init = ''
                results_primary_keys = ''
                for each_primary_key in primary_key:
                    filter_list_init += CodeBlockTemplate.multi_primary_key_filter.format(
                        primary_key=each_primary_key
                    )
                    results_primary_keys += CodeBlockTemplate.multi_primary_key_result.format(
                        primary_key=each_primary_key
                    )
                if self.flask:
                    update = FileTemplate.flask_update_template_physical.format(
                        filter_list_init=filter_list_init,
                        results_primary_keys=results_primary_keys
                    )
                else:
                    update = FileTemplate.update_template_physical.format(
                        filter_list_init=filter_list_init,
                        results_primary_keys=results_primary_keys
                    )

                # # 列表添加模块
                # add_list_column_init = ''
                # add_list_business_key_init = ''
                # for column in table['columns'].values():
                #     if column['name'] == primary_key and business_key != primary_key:
                #         continue
                #     if column['name'] == table['logical_delete_column']:
                #         continue
                #
                #     if column['name'] not in table['rsa_columns']:
                #         # 字段不需要加密
                #         if business_key != column['name']:
                #             # 当前字段不是业务主键
                #             text = CodeBlockTemplate.add_list_column_init.format(column=column['name'])
                #         else:
                #             # 当前字段是业务主键
                #             if table['business_key_column'].get('rule'):
                #                 # 是业务主键且有生成规则
                #                 text = CodeBlockTemplate.business_key_add.format(column=column['name'])
                #                 add_list_business_key_init = CodeBlockTemplate.add_list_business_key_init.format(
                #                     business_key=column['name'],
                #                     rule=table['business_key_column']['rule']
                #                 )
                #             else:
                #                 # 是业务主键但是没有生成规则
                #                 text = CodeBlockTemplate.add_list_column_init.format(column=column['name'])
                #
                #     else:
                #         # 字段需要加密
                #         text = CodeBlockTemplate.add_list_rsa_add.format(column=column['name'])
                #
                #     add_list_column_init += text
                #
                # # 拼接返回字段
                # added_record_primary_keys = ''
                # if len(table['primary_key_columns']) > 1:
                #     # 属于复合主键
                #     for each_primary_key in primary_key:
                #         added_record_primary_keys += CodeBlockTemplate.multi_primary_key_add_list_result_detail_keys.format(
                #             primary_key=each_primary_key
                #         )
                # else:
                #     # 不属于复合主键
                #     added_record_primary_keys += CodeBlockTemplate.multi_primary_key_add_list_result_detail_keys.format(
                #         primary_key=business_key if business_key else primary_key
                #     )
                #
                # add_list = FileTemplate.add_list_template.format(
                #     parent_model=parent_model,
                #     add_list_business_key_init=add_list_business_key_init,
                #     add_list_column_init=add_list_column_init,
                #     added_record_primary_keys=added_record_primary_keys
                # )

                # save into 'codes'
                file_name = little_camel_case_str + 'Controller'
                codes[file_name] = basic + add + get + delete + update

            # generate files
            print("Info: ", 'Generating __init__...')
            inti_file = os.path.join(controller_dir, '__init__.py')
            with open(inti_file, 'w', encoding='utf-8') as fw:
                if self.flask:
                    fw.write(FileTemplate.init_flask_template)
                else:
                    fw.write(FileTemplate.init_template.format(databaseUrl=self.databaseUrl))

            print("Info: ", '__init__ generated successfully')
            for file_name, code in codes.items():
                print("Info: ", 'Generating {}...'.format(file_name))
                m_file = os.path.join(controller_dir, file_name + '.py')
                with open(m_file, 'w', encoding='utf-8') as fw:
                    fw.write(code)
                print("Info: ", '{} generated successfully'.format(file_name))

        except Exception as e:
            print("Error: ", e)

    @classmethod
    def static_generate(cls, target_dir, source_dir):
        """
        1 copy the static resource to target project directory;
        2 you can put these static resource  into "static" directory,such as "dockerfile" and some
         common tools(or function) that you will use in your target project;
        3 some resource we need has already copied into default static directory;
        :param target_dir: Target path of the file
        :param source_dir: Source path of the file
        :param session_id: The ID of User
        :return: None
        """

        try:
            # 判断目标路径状态
            if not os.path.exists(target_dir):
                os.makedirs(target_dir)

            # 拷贝
            if os.path.exists(source_dir):
                for root, dirs, files in os.walk(source_dir):
                    if os.path.basename(root) == "__pycache__":
                        continue
                    for file in files:
                        # 源文件路径
                        src_file = os.path.join(root, file)
                        # 目标文件路径
                        target_file = target_dir + root.replace(source_dir, '')
                        if not os.path.exists(target_file):
                            os.makedirs(target_file)
                        # 拷贝
                        shutil.copy(src_file, target_file)
                        print("Info: ", "The file '{}' has been copied to '{}'".format(src_file, target_file))

        except Exception as e:
            print("Error: ", e)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# file:main.py
# author:Itsuka
# datetime:2021/8/26 14:33
# software: PyCharm

"""
    this is function description
"""

import os

from .codegenerator import CodeGenerator


def main(table_dict, settings):
    """
    Generate Controller code
    :return: None
    """

    project_dir = settings.PROJECT_DIR

    # create the controller file
    os.makedirs(controller_dir := os.path.join(project_dir, 'controller'), exist_ok=True)

    generator = CodeGenerator(table_dict)
    generator.controller_codegen(
        controller_dir=controller_dir,
    )
    return



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# file:codeblocktemplate.py
# author:Itsuka
# datetime:2021/8/26 10:32
# software: PyCharm

"""
    provide code block template here
"""


class CodeBlockTemplate(object):

    flask_imports = """
import datetime
import math

from ..models import db
from ..models.{model_name} import {parent_model}
"""

    imports = """
import datetime
import math

from . import session
from ..models.{model_name} import {parent_model}
"""

    add_column_init = """{column}=kwargs.get('{column}'),
                """

    rsa_add = """{column}=RSAEncryptionDecryption.encrypt(kwargs.get('{column}')) if kwargs.get('{column}') else None,
                """

    business_key_add = """{column}={column},
                """

    get_filter_num = """if kwargs.get('{column}') is not None:
                    filter_list.append(cls.{column} == kwargs.get('{column}'))
                """

    get_filter_str = """if kwargs.get('{column}'):
                    filter_list.append(cls.{column} == kwargs.get('{column}'))
                """

    get_filer_list_logic = 'cls.{logical_delete_mark} == 0'

    rsa_get_filter_num = """if kwargs.get('{column}') is not None:
                    filter_list.append(cls.{column} == RSAEncryptionDecryption.encrypt(kwargs.get('{column}')))
                """

    rsa_get_filter_str = """if kwargs.get('{column}'):
                    filter_list.append(cls.{column} == RSAEncryptionDecryption.encrypt(kwargs.get('{column}')))
                """

    rsa_update = """if kwargs.get('{column}'):
                kwargs['{column}'] = RSAEncryptionDecryption.encrypt(kwargs['{column}'])
            """

    business_key_init = """from utils.generate_id import GenerateID
        {business_key} = GenerateID.{rule}()
        """

    add_list_column_init = """{column}=param_dict.get('{column}'),
                """

    add_list_business_key_init = """from utils.generate_id import GenerateID
            {business_key} = GenerateID.{rule}()
            """

    add_list_rsa_add = """{column}=RSAEncryptionDecryption.encrypt(param_dict.get('{column}')),
                """

    single_primary_key_get_filter = """if kwargs.get('{primary_key}'):
                filter_list.append(cls.{primary_key} == kwargs['{primary_key}'])
            else:
                {get_filter_list}
"""

    delete_filter = """if kwargs.get('{primary_key}'):
                primary_key_list = []
                for primary_key in str(kwargs.get('{primary_key}')).replace(' ', '').split(','):
                    primary_key_list.append(cls.{primary_key} == primary_key)
                filter_list.append(or_(*primary_key_list))
                
            else:
                {delete_filter_list}"""

    multi_primary_key_filter = """filter_list.append(cls.{primary_key} == kwargs.get('{primary_key}'))
            """

    results_primary_keys = """'{primary_key}': []"""

    multi_primary_key_result = """'{primary_key}': res.first().{primary_key},
                """

    add_result_primary_key = """'{primary_key}': model.{primary_key},
                """

    # add_list_result_primary_key = """'{primary_key}': [],"""
    #
    # multi_primary_key_add_list_result = """'records': []"""

    multi_primary_key_add_list_result_detail_keys = """added_record['{primary_key}'] = model.{primary_key}
                """

    single_primary_key_result_append = """for query_model in res.all():
                results['{primary_key}'].append(query_model.{primary_key})
"""

    multi_get_filter_num = """if kwargs.get('{column}') is not None:
                filter_list.append(cls.{column} == kwargs.get('{column}'))
            """

    multi_get_filter_str = """if kwargs.get('{column}'):
                filter_list.append(cls.{column} == kwargs.get('{column}'))
            """

    multi_rsa_get_filter_num = """if kwargs.get('{column}') is not None:
                filter_list.append(cls.{column} == RSAEncryptionDecryption.encrypt(kwargs.get('{column}')))
            """

    multi_rsa_get_filter_str = """if kwargs.get('{column}'):
                filter_list.append(cls.{column} == RSAEncryptionDecryption.encrypt(kwargs.get('{column}')))
            """



import sys

from setuptools import setup, find_packages
# from setuptools.command.test import test as TestCommand
#
# import sqlacodegen
#
#
# class PyTest(TestCommand):
#     def finalize_options(self):
#         TestCommand.finalize_options(self)
#         self.test_args = []
#         self.test_suite = True
#
#     def run_tests(self):
#         import pytest
#         errno = pytest.main(self.test_args)
#         sys.exit(errno)
import codegen

extra_requirements = ()
if sys.version_info < (2, 7):
    extra_requirements = ('argparse',)

setup(
    name='sqlalchemy-codegen',
    description='Automatic generate model layer and controller layer code for SQLAlchemy with Flask support',
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    version=codegen.version,
    author='iDeal-ncepu',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Topic :: Database',
        'Topic :: Software Development :: Code Generators',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords=['sqlalchemy','flask-sqlalchemy', 'sqlacodegen', 'flask'],
    license='MIT',
    packages=find_packages(exclude=['tests']),
    package_data={'': ['*.json', '*.conf']},
    install_requires=(
        'SQLAlchemy >= 0.6.0',
        'inflect >= 0.2.0',
        'pymysql>= 1.0.2',
        'loguru >= 0.6.0',
        'sqlacodegen>=2.3.0',
        'greenlet>=1.1.2',
    ) + extra_requirements,
    # tests_require=['pytest', 'pytest-pep8'],
    # cmdclass={'test': PyTest},
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'sqlalchemy-codegen=codegen.main:main'
        ]
    }
)



# coding:utf-8

class RET:
    OK = "2000"

    NODATA = "2001"
    DATAEXIST = "2002"
    DATAERR = "2003"
    PARAMERR = "4000"
    SESSIONERR = "4001"
    LOGINERR = "4002"
    USERERR = "4003"
    ROLEERR = "4004"
    PWDERR = "4005"
    REQERR = "4006"
    IPERR = "4007"
    USERLOCKERR = "4008"
    INTERNALERR = "5000"
    DBERR = "5001"
    THIRDERR = "5002"
    IOERR = "5003"
    UNKOWNERR = "5004"


error_map = {
    RET.OK: u"成功",

    RET.NODATA: u"无数据",
    RET.DATAEXIST: u"数据已存在",
    RET.DATAERR: u"数据错误",
    RET.PARAMERR: u"参数错误",
    RET.SESSIONERR: u"用户未登录",
    RET.LOGINERR: u"用户登录失败",
    RET.USERERR: u"用户不存在或未激活",
    RET.ROLEERR: u"用户身份错误",
    RET.PWDERR: u"密码错误",
    RET.REQERR: u"非法请求或请求次数受限",
    RET.IPERR: u"IP受限",
    RET.USERLOCKERR: u"账号被锁定",
    RET.DBERR: u"数据库查询错误",
    RET.THIRDERR: u"第三方系统错误",
    RET.IOERR: u"文件读写错误",
    RET.UNKOWNERR: u"未知错误",
}
error_map_EN = {
    RET.OK: "successfully!",
    RET.NODATA: "no data!",
    RET.DATAEXIST: "data exist!",
    RET.DATAERR: "data error!",
    RET.PARAMERR: "parameter error!",
    RET.SESSIONERR: "user not logged in!",
    RET.LOGINERR: "user login failed!",
    RET.USERERR: "user does not exist or not activation! ",
    RET.ROLEERR: "user role error!",
    RET.PWDERR: "password error!",
    RET.REQERR: "illegal request or limited number of requests",
    RET.IPERR: "IP restricted!",

    RET.DBERR: "database query error",
    RET.THIRDERR: "third party system error",
    RET.IOERR: "file read or write error!",
    RET.UNKOWNERR: "unknown error!",
}
